
#include <linux/tty.h>
#include <linux/smp_lock.h>
#include <linux/sched.h>
#include <linux/file.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/list.h>
#include <linux/idr.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include <linux/miscdevice.h>

#include <net/sock.h>
#include <net/netlink.h>

#include "sniffy.h"
#include "isniffy.h"
#include "autoconf.h"

static struct hlist_head *sniffy_term_hash;
static struct hlist_head *sniffy_term_pid_hash;
static DEFINE_RWLOCK(sniffy_term_lock);

static struct sock *sniffy_sock;

/* forward declarations */
static struct sniffy_term* sniffy_hash_find(int idx);
static int sniffy_process(struct sniffy_term *term, int type, const unsigned char *buf, size_t len);
static int sniffy_detach(int idx, int full, int except);
static int sniffy_pid_attach(pid_t pid, int idx);
static int sniffy_pid_detach(pid_t pid, int idx);


static struct tty_struct *get_tty(int idx)
{
	char buf[sizeof(SF_DEVPTS_PATH"/000000000000")];
	struct nameidata nd;
	struct dentry *d;
	struct fs_struct *fs;

	BUG_ON(NULL == (fs = current->fs));

	DEBUG2("trying to get idx %d\n", idx);
	sprintf(buf, SF_DEVPTS_PATH"/%d", idx);

	if (vfs_path_lookup(fs->root.dentry, fs->root.mnt, 
			buf, LOOKUP_FOLLOW, &nd))
		return NULL;

	d = nd.path.dentry;
	DEBUG2("got inode %lu\n", d->d_inode->i_ino);

	return d->d_inode->i_private;
}

static void ldisc_close_pty(struct tty_struct *tty)
{
	struct sniffy_term *term;

	read_lock(&sniffy_term_lock);

	term = sniffy_hash_find(tty->index);
	BUG_ON(!term);

	term->ld_close(tty);
	DEBUG2("pty %d closed\n", term->idx);

	sniffy_detach(term->idx, 1, 0);

	read_unlock(&sniffy_term_lock);
}

static ssize_t ldisc_write_pty(struct tty_struct *tty, struct file *file,
			    const unsigned char *buf, size_t nr)
{
	struct sniffy_term *term;
	int type = SNIFFY_MSG_DATA_SLAVE, rc;
	ldisc_write_t ld_write_slave;
	ldisc_write_t ld_write_master;
	struct tty_struct *tty_master, *tty_slave;

	read_lock(&sniffy_term_lock);

	term = sniffy_hash_find(tty->index);
	BUG_ON(!term);

	if (tty == term->tty_master)
		type = SNIFFY_MSG_DATA_MASTER;

	rc = sniffy_process(term, type, buf, nr);

	ld_write_slave = term->ld_write_slave;
	ld_write_master = term->ld_write_master;
	tty_slave = term->tty_slave;
	tty_master = term->tty_master;

	read_unlock(&sniffy_term_lock);

	if (tty == tty_slave)
		return ld_write_slave(tty, file, buf, nr);

	if (tty == tty_master)
		return ld_write_master(tty, file, buf, nr);

	BUG();
	return 0;
}

static int sniffy_patch_pty(struct sniffy_term *term)
{
	struct tty_ldisc *ld_slave;
	struct tty_ldisc *ld_master;

	if (NULL == (term->tty_slave = get_tty(term->idx))) {
		DEBUG2("failed to get pty %d\n", term->idx);
		return 1;
	}

	term->tty_master = term->tty_slave->link;

	DEBUG2("patching master %d\n", term->idx);

	ld_master = tty_ldisc_ref_wait(term->tty_master);
	term->ld_write_master = ld_master->write;
	ld_master->write = ldisc_write_pty;
	term->ld_close = ld_master->close;
	ld_master->close = ldisc_close_pty;
	tty_ldisc_deref(ld_master);

	DEBUG2("patching slave %d\n", term->idx);

	ld_slave = tty_ldisc_ref_wait(term->tty_slave);
	term->ld_write_slave = ld_slave->write;
	ld_slave->write = ldisc_write_pty;
	tty_ldisc_deref(ld_slave);

	return 0;
}

static int sniffy_restore_pty(struct sniffy_term *term, int except)
{
        struct tty_ldisc *ld_slave;
        struct tty_ldisc *ld_master;

	DEBUG2("restoring master %d\n", term->idx);
	ld_master = except ? tty_ldisc_ref_wait(term->tty_master) : &term->tty_master->ldisc;
	ld_master->write = term->ld_write_master;
	ld_master->close = term->ld_close;

	if (except)
		tty_ldisc_deref(ld_master);

	DEBUG2("restoring slave %d\n", term->idx);
	ld_slave = except ? tty_ldisc_ref_wait(term->tty_slave) : &term->tty_slave->ldisc;
	ld_slave->write = term->ld_write_slave;

	if (except)
		tty_ldisc_deref(ld_slave);

	return 0;
}

#define SNIFFY_HASH_SIZE 32
static inline int sniffy_hash(int idx)
{
        return idx % SNIFFY_HASH_SIZE;
}

static int sniffy_hash_add(struct sniffy_term *term)
{
	struct hlist_head *head;

	DEBUG2("pty %d\n", term->idx);

	head = &sniffy_term_hash[sniffy_hash(term->idx)];
	hlist_add_head(&term->node, head);

	return 0;
}

static int sniffy_hash_del(struct sniffy_term *term)
{
	struct hlist_head *head;
	struct hlist_node *node;
	struct sniffy_term *t = NULL;

	DEBUG2("pty %d\n", term->idx);

	head = &sniffy_term_hash[sniffy_hash(term->idx)];
	hlist_for_each_entry(t, node, head, node) {
		if (t == term)
			break;
	}

	hlist_del(&term->node);

	return 0;
}

static struct sniffy_term* sniffy_hash_find(int idx)
{
	struct hlist_head *head;
	struct hlist_node *node;
	struct sniffy_term *term;

	head = &sniffy_term_hash[sniffy_hash(idx)];
	hlist_for_each_entry(term, node, head, node) {
		if (term->idx == idx)
			return term;
	}

	return NULL;
}

static struct sk_buff *sniffy_make_msg(int pid, int seq, int type, int done,
                                 int multi, int size)
{
	struct sk_buff  *skb;
	struct nlmsghdr *nlh;
	int len = NLMSG_SPACE(size);
	int flags = multi ? NLM_F_MULTI : 0;
	int t = done ? NLMSG_DONE  : type;

	skb = alloc_skb(len, GFP_KERNEL);
	if (!skb)
		return NULL;

	nlh = NLMSG_PUT(skb, pid, seq, t, size);
	nlh->nlmsg_flags = flags;
	return skb;

nlmsg_failure: /* Used by NLMSG_PUT */
	if (skb)
		kfree_skb(skb);
	return NULL;
}

static int sniffy_process(struct sniffy_term *term, int type, const unsigned char *buf, size_t len)
{
	struct sk_buff *skb;
	struct sniffy_msg_data dmsg, *dmsgp;
	void *data;
	struct nlmsghdr *nlh;

	dmsg.cmsg.idx = term->idx;
	dmsg.type = type;

	skb = sniffy_make_msg(0, term->seq, SNIFFY_CMD_DATA, 0, 0, sizeof(dmsg) + len);
	nlh = nlmsg_hdr(skb);
	data = dmsgp = NLMSG_DATA(nlh);

	/* window size data */
	dmsg.ws.ws_row = term->tty_slave->winsize.ws_row;
	dmsg.ws.ws_col = term->tty_slave->winsize.ws_col;
	dmsg.ws.ws_xpixel = term->tty_slave->winsize.ws_xpixel;
	dmsg.ws.ws_ypixel = term->tty_slave->winsize.ws_ypixel;

	memcpy(data, &dmsg, sizeof(dmsg));
	memcpy(dmsgp->data, buf, len);
	dmsgp->len = len;

	DEBUG2("got data len %u\n", len);
	return netlink_broadcast(sniffy_sock, skb, 0, term->group, GFP_ATOMIC);
}

static DEFINE_IDR(allocated_groups);
static DEFINE_MUTEX(allocated_groups_lock);

static unsigned int sniffy_get_group(void)
{
	int idr_ret;
	unsigned int group;

	mutex_lock(&allocated_groups_lock);

	if (!idr_pre_get(&allocated_groups, GFP_KERNEL)) {
		mutex_unlock(&allocated_groups_lock);
		return -ENOMEM;
	}

	idr_ret = idr_get_new(&allocated_groups, NULL, &group);
	if (idr_ret < 0) {
		mutex_unlock(&allocated_groups_lock);
		if (idr_ret == -EAGAIN)
			return -ENOMEM;
		return -EIO;
	}

	/* group zero is special for netlink, 
	   so we start from 1 */
	group++;

	DEBUG2("got group %u\n", group);

	mutex_unlock(&allocated_groups_lock);

	return group;
}


static void sniffy_kill_group(int group)
{
	mutex_lock(&allocated_groups_lock);
	idr_remove(&allocated_groups, --group);
	mutex_unlock(&allocated_groups_lock);
}

static int sniffy_attach(struct nlmsghdr *nlh, struct sniffy_msg_common* cmsg, u32 pid)
{
	struct sk_buff *skb;
	struct sniffy_msg_group gmsg;
	struct sniffy_term *term;
	void *data;
	int rc = 0;

	DEBUG2("idx = %u, pid %d\n", cmsg->idx, pid);

	write_lock(&sniffy_term_lock);

	if ((rc = sniffy_pid_attach(pid, cmsg->idx)))
		goto out;

	if (!(term = sniffy_hash_find(cmsg->idx))) {

		if (!(term = kmalloc(sizeof(*term), GFP_KERNEL))) {
			rc = -ENOMEM;
			sniffy_pid_detach(pid, cmsg->idx);
			goto out;
		}

		memset(term, 0, sizeof(*term));
		term->idx = cmsg->idx;
		term->group = sniffy_get_group();

		if ((rc = sniffy_patch_pty(term))) {
			kfree(term);
			sniffy_pid_detach(pid, cmsg->idx);
			goto out;
		}

		sniffy_hash_add(term);
	}

	atomic_inc(&term->refcnt);

	gmsg.cmsg = *cmsg;
	gmsg.group = term->group;

	skb = sniffy_make_msg(pid, nlh->nlmsg_seq, SNIFFY_CMD_GROUP, 
				0, 0, sizeof(gmsg));
	nlh = nlmsg_hdr(skb);
	data = NLMSG_DATA(nlh);
	memcpy(data, &gmsg, sizeof(gmsg));

	term->seq = nlh->nlmsg_seq + 1;

	netlink_unicast(sniffy_sock, skb, pid, 0);

out:
	write_unlock(&sniffy_term_lock);
	return rc;
}

static int __sniffy_detach(int idx, int full, int except)
{
	struct sniffy_term *term;
	struct sniffy_msg_common cmsg;
	void *data;
	struct sk_buff *skb;
	struct nlmsghdr *nlh;
	int rc = 0;

	DEBUG2("idx = %u, full = %d, except = %d\n", idx, full, except);

	if (!(term = sniffy_hash_find(idx))) {
		rc = -EINVAL;
		goto out;
	}

	if (!full && !atomic_dec_and_test(&term->refcnt))
		goto out;

	sniffy_restore_pty(term, except);
	sniffy_hash_del(term);
	
	/* notify listenners we are detaching */
	cmsg.idx = term->idx;
	skb = sniffy_make_msg(0, term->seq, SNIFFY_CMD_DETACH, 
				0, 0, sizeof(cmsg));
	nlh = nlmsg_hdr(skb);
	data = NLMSG_DATA(nlh);

	memcpy(data, &cmsg, sizeof(cmsg));
	netlink_broadcast(sniffy_sock, skb, 0, term->group, GFP_ATOMIC);
	DEBUG2("detached idx = %u\n", idx);

	sniffy_kill_group(term->group);
	kfree(term);

out:
	sniffy_pid_detach(except ? current->pid : -1, idx);
	return 0;
}

static int sniffy_detach(int idx, int full, int except)
{
	int ret;

	write_lock(&sniffy_term_lock);
	ret = __sniffy_detach(idx, full, except);
	write_unlock(&sniffy_term_lock);

	return ret;
}

static int sniffy_receive_msg(struct sk_buff *skb, struct nlmsghdr *nlh)
{
	u16 msg_type = nlh->nlmsg_type;
	u32 pid;
	void *data;
	struct sniffy_msg_common *cmsg;

        data = NLMSG_DATA(nlh);
	cmsg = (struct sniffy_msg_common*) data;
	pid = NETLINK_CREDS(skb)->pid;

	switch(msg_type) {
	case SNIFFY_CMD_ATTACH:
		sniffy_attach(nlh, cmsg, pid);
		break;
	case SNIFFY_CMD_DETACH:
		sniffy_detach(cmsg->idx, 0, 1);
		break;
	}

	return 0;
}

static void sniffy_receive_skb(struct sk_buff  *skb)
{
	int             err;
	struct nlmsghdr *nlh;
	u32             rlen;

	while (skb->len >= NLMSG_SPACE(0)) {

		nlh = nlmsg_hdr(skb);
		if (nlh->nlmsg_len < sizeof(*nlh) || skb->len < nlh->nlmsg_len)
			return;

		rlen = NLMSG_ALIGN(nlh->nlmsg_len);
		if (rlen > skb->len)
			rlen = skb->len;

		if ((err = sniffy_receive_msg(skb, nlh))) {
			netlink_ack(skb, nlh, err);
		} else if (nlh->nlmsg_flags & NLM_F_ACK)
			netlink_ack(skb, nlh, 0);

		skb_pull(skb, rlen);
	}
}       

static int proc_show(struct seq_file *seq, void *v)
{
	int i;
	struct hlist_head *head;
	struct hlist_node *node;
	struct sniffy_term *term;
	struct sniffy_term_pid *term_pid;

	read_lock(&sniffy_term_lock);

	seq_printf(seq, "%-5s %-10s %-10s\n", "pty", "group", "ref");

	for(i = 0; i < SNIFFY_HASH_SIZE; i++) {
		head = &sniffy_term_hash[i];
		hlist_for_each_entry(term, node, head, node)
			seq_printf(seq, "%-5u %-10u %-10u\n", 
					term->idx, 
					term->group, 
					atomic_read(&term->refcnt));
        }

	seq_printf(seq, "\n");
	seq_printf(seq, "%-5s %s\n", "pid", "idx");

	for(i = 0; i < SNIFFY_HASH_SIZE; i++) {
		head = &sniffy_term_pid_hash[i];
		hlist_for_each_entry(term_pid, node, head, node) {
			int j, k;

			seq_printf(seq, "%-5d ", term_pid->pid);

			for(j = 0, k = 0; j < term_pid->idxcnt; j++) {

				if (term_pid->idx[j] == -1)
					continue;

				if (k++)
					seq_printf(seq, ",%d", term_pid->idx[j]);
				else
					seq_printf(seq, "%d", term_pid->idx[j]);
			}
			seq_printf(seq, "\n");
		}
			
        }

	read_unlock(&sniffy_term_lock);
	return 0;
}

static int proc_seq_open(struct inode *inode, struct file *file)
{
	return single_open(file, proc_show, NULL);
}

static struct sniffy_term_pid* sniffy_pid_find(pid_t pid)
{
	struct hlist_head *head;
	struct hlist_node *node;
	struct sniffy_term_pid *term_pid;

	head = &sniffy_term_pid_hash[sniffy_hash(pid)];
	hlist_for_each_entry(term_pid, node, head, node) {
		if (term_pid->pid == pid)
			return term_pid;
	}

	return NULL;
}

static int sniffy_pid_init(pid_t pid)
{
	struct sniffy_term_pid *term_pid;
	struct hlist_head *head;
	int i;

	write_lock(&sniffy_term_lock);

	if (sniffy_pid_find(pid))
		return -EINVAL;

	if (!(term_pid = kmalloc(sizeof(*term_pid), GFP_KERNEL)))
		return -ENOMEM;

	memset(term_pid, 0, sizeof(*term_pid));
	term_pid->pid = pid;

	if (!(term_pid->idx = kmalloc(SF_PID_MAXIDX * sizeof(int), GFP_KERNEL))) {
		kfree(term_pid);
		return -ENOMEM;
	}

	for(i = 0; i < SF_PID_MAXIDX; i++)
		term_pid->idx[i] = -1;

	head = &sniffy_term_pid_hash[sniffy_hash(pid)];
	hlist_add_head(&term_pid->node, head);

	write_unlock(&sniffy_term_lock);
	return 0;
}

static int sniffy_pid_exit(pid_t pid)
{
	struct sniffy_term_pid *term_pid;
	int i, ret = -EINVAL;

	write_lock(&sniffy_term_lock);

	if (NULL == (term_pid = sniffy_pid_find(pid)))
		goto out;

	for(i = 0; i < term_pid->idxcnt; i++) {
		struct sniffy_term *term;
		int idx = term_pid->idx[i];

		if (-1 == idx)
			continue;

		if (NULL == (term = sniffy_hash_find(term_pid->idx[i]))) {
			printk("sniffy_pid_exit - inconsistency\n");
			continue;
		}

		__sniffy_detach(term_pid->idx[i], 0, 1);
	}

        hlist_del(&term_pid->node);
	kfree(term_pid->idx);
	kfree(term_pid);
	ret = 0;

out:
	write_unlock(&sniffy_term_lock);
	return ret;
}

static int sniffy_pid_attach(pid_t pid, int idx)
{
	struct sniffy_term_pid *term_pid;
	int i;

	if (NULL == (term_pid = sniffy_pid_find(pid)))
		return -EINVAL;

	for(i = 0; i < term_pid->idxcnt; i++)
		if (term_pid->idx[i] == -1)
			break;

	if (i == term_pid->idxcnt) {

		if (term_pid->idxcnt < SF_PID_MAXIDX)
			term_pid->idxcnt++;
		else
			return -EINVAL;
	}

	term_pid->idx[i] = idx;
	return 0;
}

static int __sniffy_pid_detach(struct sniffy_term_pid *term_pid, int idx)
{
	int i, ret = 0;

	DEBUG2("pid = %d, idx = %d\n", term_pid->pid, idx);

	for(i = 0; i < term_pid->idxcnt; i++)
		if (term_pid->idx[i] == idx)
			break;

	/* term_pid record not found */
	if (i == term_pid->idxcnt) {
		ret = -EINVAL;
		goto out;
	}

	term_pid->idx[i] = -1;
	if ((i + 1) == term_pid->idxcnt)
		term_pid->idxcnt--;

out:
	return ret;
}

static int sniffy_pid_detach(pid_t pid, int idx)
{
	struct sniffy_term_pid *term_pid;
	int ret = 0;

	if (pid != -1) {

		if (NULL == (term_pid = sniffy_pid_find(pid))) {
			ret = -EINVAL;
			goto out;
		}

		ret = __sniffy_pid_detach(term_pid, idx);

	} else {

		struct hlist_head *head;
		struct hlist_node *node;
		int i;

		for(i = 0; i < SNIFFY_HASH_SIZE; i++) {
			head = &sniffy_term_pid_hash[i];
			hlist_for_each_entry(term_pid, node, head, node)
				__sniffy_pid_detach(term_pid, idx);
		}
        }

out:
	return ret;
}

static int ctrl_task_init(struct inode *inode, struct file *file)
{
	pid_t pid = current->pid;
	DEBUG2("task with pid %d created\n", pid);
	return sniffy_pid_init(pid);
}

static int ctrl_task_exit(struct file *file, fl_owner_t id)
{
	pid_t pid = current->pid;
	DEBUG2("task with pid %d finished\n", pid);
	return sniffy_pid_exit(pid);
}

static const struct file_operations ctrl_fops = {
	.owner          = THIS_MODULE,
        .open           = ctrl_task_init,
        .flush          = ctrl_task_exit,
};

static struct miscdevice ctrl_dev = {
        .minor          = MISC_DYNAMIC_MINOR,
        .name           = SNIFFY_DEV,
        .fops           = &ctrl_fops,
};

static struct file_operations proc_sniffy_operations = {
	.owner      = THIS_MODULE,
	.open       = proc_seq_open,
	.read       = seq_read,
	.llseek     = seq_lseek,
	.release    = single_release,
};

int sniffy_debug = 0;
module_param_named(debug, sniffy_debug, int, 0444);
MODULE_PARM_DESC(debug, "produce debug output");

int sniffy_proc = 0;
module_param_named(proc, sniffy_proc, int, 0444);
MODULE_PARM_DESC(debug, "create proc file /proc/sniffy");

static int __init sniffy_init(void)
{
	int i, ret = -ENOMEM;

	if (NULL == (sniffy_term_hash = kmalloc(SNIFFY_HASH_SIZE * sizeof(struct hlist_head), GFP_KERNEL)))
		return ret;

	if (NULL == (sniffy_term_pid_hash = kmalloc(SNIFFY_HASH_SIZE * sizeof(struct hlist_head), GFP_KERNEL)))
		goto out1;

	for(i = 0; i < SNIFFY_HASH_SIZE;  i++) {
		INIT_HLIST_HEAD(&sniffy_term_hash[i]);
		INIT_HLIST_HEAD(&sniffy_term_pid_hash[i]);
	}

	sniffy_sock = netlink_kernel_create(&init_net, 
				NETLINK_SNIFFY, NR_UNIX98_PTY_MAX,
				sniffy_receive_skb, NULL, THIS_MODULE);

	if (!sniffy_sock) {
		printk("sniffy - netlink_kernel_create failed\n");
		ret = -ENOENT;
		goto out2;
	}

	if (sniffy_proc)
		proc_create("sniffy", 0, NULL, &proc_sniffy_operations);

	if (misc_register(&ctrl_dev))
		goto out3;

	printk(KERN_INFO "sniffy v"SNIFFY_VER" loaded (olsajiri@gmail.com)\n");
	return 0;

out3:
	if (sniffy_proc)
		remove_proc_entry("sniffy", NULL);
out2:
	kfree(sniffy_term_pid_hash);
out1:
	kfree(sniffy_term_hash);
	return ret;
}

static void __exit sniffy_cleanup(void)
{
	netlink_kernel_release(sniffy_sock);

	/* We dont need to detach any terminal, since 
	   module can be unloaded only if there's no reference to it.
	   So in this point every sniffed terminal should be detached,
	   and the hash arrays are the only job left. */
	kfree(sniffy_term_hash);
	kfree(sniffy_term_pid_hash);

	if (sniffy_proc)
		remove_proc_entry("sniffy", NULL);

	misc_deregister(&ctrl_dev);

	printk(KERN_INFO "sniffy v"SNIFFY_VER" unloaded (olsajiri@gmail.com)\n");
}

module_init(sniffy_init);
module_exit(sniffy_cleanup);

MODULE_LICENSE("GPL");
