
#include <string.h>
#include <stdlib.h>
#include <getopt.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ftw.h>

#include "config.h"


static const struct sf_config default_config = {
	.sniffy_dir         = "/tmp/sniffy",
	.trace_all          = 1,
};

static int config_trace(struct sf_config *cfg, char *lv, char *rv)
{
	if (!strncmp(lv, "dir", sizeof("dir"))) {
		cfg->sniffy_dir = rv;
		return 0;
	}

	/* no trace at all */
	if (!strncmp(lv, "none", sizeof("none"))) {
		cfg->trace_all = 0;
		return 0;
	}

	return 0;
}

static int config_x(struct sf_config *cfg, char *lv, char *rv)
{
#ifdef CONFIG_SNIFFY_X_RESIZE
	if (!strncmp(lv, "resize", sizeof("resize")) && rv) {
		cfg->resize = atoi(rv);
		return 0;
	}
#endif
	return 0;
}

#define CONFIG_TRACE     1
#define CONFIG_X         2

static int config_parse(struct sf_config *cfg, int type, char *str)
{
	char *ass, *save;

	for(;(ass = strtok_r(str, ",", &save)) ;str = NULL) {
		char *lv, *rv;
		int ret;

		lv = strtok(ass, "=");
		rv = strtok(NULL, "=");

		switch(type) {
		case CONFIG_TRACE: 
			ret = config_trace(cfg, lv, rv); 
			break;
		case CONFIG_X: 
			ret = config_x(cfg, lv, rv); 
			break;
		default:
			printf("unknown type");
			return -1;
		}

		if (-1 == ret)
			return -1;
	}

	return 0;
}

static inline int config_terminal_trace(struct sf_config *cfg)
{
	return cfg->trace_all;
}

static int config_make_depends(struct sf_config *cfg)
{
	struct sf_terminal *term = NULL;

	/* setup term based on global settings */
	for(term = cfg->terminal; term != NULL; term = term->next)
		term->trace = config_terminal_trace(cfg);

	return 0;
}

struct sf_terminal* sf_config_terminal_find(struct sf_config *cfg, int idx)
{
	struct sf_terminal *term = NULL;

	for(term = cfg->terminal; term != NULL; term = term->next) {
		if (term->idx == idx)
			break;
	}

	return term;
}

struct sf_terminal* sf_config_terminal_add(struct sf_config *cfg, int idx)
{
	struct sf_terminal *term;

	if (NULL == (term = malloc(sizeof(*term)))) {
		perror("malloc failed");
		return NULL;
	}

	memset(term, 0, sizeof(*term));
	term->state = SF_STATE_NEW;
	term->idx   = idx;
	term->trace = config_terminal_trace(cfg);

	term->next = cfg->terminal;
	if (cfg->terminal)
		cfg->terminal->prev = term;
	cfg->terminal = term;

	if (!cfg->terminal_cnt)
		cfg->terminal_current = term;

	cfg->terminal_cnt++;
	return term;
}

int sf_config_terminal_del(struct sf_config *cfg, struct sf_terminal *term)
{
	if (cfg->terminal == term)
		cfg->terminal = term->next;
	else if (term->prev)
		term->prev->next = term->next;

	if (term->next)
		term->next->prev = term->prev;

	cfg->terminal_cnt--;
	free(term);

	if (cfg->terminal_current == term)
		cfg->terminal_current = cfg->terminal;
	else if (!cfg->terminal_cnt)
		cfg->terminal_current = NULL;
	
	return 0;
}

static struct sf_config *scfg = NULL;

static int ftw_cb(const char *fpath, const struct stat *sb, int typeflag)
{
	int idx;
	if (S_ISDIR(sb->st_mode))
		return 0;

	idx = (sb->st_rdev & 0xff);
	if (NULL == sf_config_terminal_add(scfg, idx))
		return -1;

	SF_LOG_PRINT2(scfg, "got idx %d\n", idx);
	return 0;
}


static int config_pty_all(struct sf_config *cfg)
{
	scfg = cfg;
	return ftw(SF_DEVPTS_PATH, ftw_cb, 1);
}

static int config_pty(struct sf_config *cfg, char *str)
{
	char *ass, *save;

	if (!strcmp("all", str))
		return config_pty_all(cfg);

	for(;(ass = strtok_r(str, ",", &save)) ;str = NULL) {
		int idx;

		if ((idx = strtol(ass, (char **) NULL, 10)) < 0)
			continue;

		if (NULL == sf_config_terminal_add(cfg, idx))
			return -1;
	}

	return cfg->terminal_cnt > 0 ? 0 : -1;
}

void usage(void)
{
	printf("usage: sniffy [-pvtrdcfx]\n");
	printf("    -p, --pty all|idx1[,idx2 ...]   watch given ptys or all\n");
	printf("    -t, --trace dir=DIR|none        trace options\n");
	printf("    -d, --daemon                    run as daemon\n");
	printf("    -c, --client                    connect to running daemon\n");
	printf("    -x, --x                         enable X features\n");
	printf("    -t, --replay tracefile          replay trace file\n");
	printf("    -v, --verbose                   display verbose/debug info\n");
	printf("    -h, --help                      display help\n");
	printf("\n");

        exit(1);

}

int sf_config_cmdline(struct sf_config *cfg, int argc, char **argv)
{
	/* setup default configuration settings */
	memset(cfg, 0, sizeof(*cfg));
	*cfg = default_config;
	cfg->tty_out = stdout;

	while (1) {
		int c;
		int option_index = 0;
		static struct option long_options[] = {
			{"pty", required_argument, 0, 'p'},
			{"verbose", no_argument , 0, 'v'},
			{"trace", no_argument , 0, 't'},
			{"replay", required_argument, 0, 'r'},
			{"daemon", no_argument, 0, 'd'},
			{"client", no_argument, 0, 'c'},
			{"config", no_argument, 0, 'f'},
			{"x", no_argument, 0, 'x'},
			{"help", no_argument, 0, 'h'},
			{0, 0, 0, 0}
		};

		c = getopt_long(argc, argv, "p:vt:r:xdch", long_options, 
				&option_index);

		if (c == -1)
			break;

		switch (c) {
		/* attach PTY */
		case 'p':
			if (-1 == config_pty(cfg, optarg))
				return -1;
			break;

		case 'v':
			cfg->verbose = 1;
			break;

		case 't':
			if (-1 == config_parse(cfg, CONFIG_TRACE, optarg))
				return -1;
			break;

		case 'r':
			cfg->replay = 1;
			cfg->replay_filename = optarg;
			break;

		case 'd':
			cfg->daemon = 1;
			break;

		case 'c':
			cfg->client = 1;
			break;

		case 'h':
			usage();
			break;

		case 'x':
#ifdef CONFIG_SNIFFY_X_RESIZE
			cfg->resize = 1;
			/* XXX so far there's only resize X option enabled by -x
			if (-1 == config_parse(cfg, CONFIG_X, optarg))
				return -1;
			*/
#else
			printf("X support not builtin\n");
#endif
			break;

		default:
			break;
		}
	}

	return config_make_depends(cfg);
}
