
#ifndef SF_CONFIG_H
#define SF_CONFIG_H

#include <sys/time.h>
#include <stdarg.h>
#include <stdio.h>
#include <autoconf.h>
#include <setjmp.h>

#include <netlink/msg.h>

#ifdef CONFIG_SNIFFY_X_RESIZE
#include <X11/X.h>
#include <X11/Xlib.h>
#endif

#include <isniffy.h>

/* daemon - client communication message */
struct sf_imsg {
#define SF_IMSG_ATTACH    1
#define SF_IMSG_DETACH    2
#define SF_IMSG_DATA      3
	int cmd;
	union {
		struct {
			int idx;
#define SF_TYPE_PTY       1
#define SF_TYPE_TTY       2
#define SF_TYPE_TTYS      3
#define SF_TYPE_CONSOLE   4
			int type;
		} attach;
		int data;
	};
};

/* trace header, record */
struct sf_trace_header {
	unsigned int magic;
	unsigned long maxidx;
	struct timeval time;
};

/* if the time is zero, data continue at the next page allignment */
struct sf_trace_record {
	u_char used;
	struct timeval time;
	int type;
	struct sniffy_msg_ws ws;
	unsigned long idx;
	int len;
	char data[0];
};


struct sf_terminal {

	/* global */
	int idx;  /* terminal index */
	int group; /* group id */

	/* terminal process state */
#define SF_STATE_NEW       1
#define SF_STATE_DATA      3
#define SF_STATE_DONE      4
	int state;

	/* trace */
	int trace_fd;
	unsigned int trace_size;
	char *trace_filename;
	unsigned long trace_idx;

#ifdef CONFIG_SNIFFY_TRACE_MMAP
#define TRACE_MMAP_SIZE (10*4096)
	void *trace_mmap_ptr;
	void *trace_mmap_cur;
#endif

	/* flags */
	int trace;       /* tracing is active for terminal */

	/* statistics */
	unsigned int rx_bytes;
	unsigned int rx_count;
	unsigned int rx_bytes_pty_slave;
	unsigned int rx_bytes_pty_master;
	unsigned int rx_count_pty_slave;
	unsigned int rx_count_pty_master;

	/* input */
	int display_slave;
	int display_master;

	/* list maintenance */
	struct sf_terminal *next;
	struct sf_terminal *prev;
};

struct sf_config {

	struct nl_handle *nlh;
	FILE *tty_out;
	char *sniffy_dir;

	/* input descriptor */
	int input_fd;

	/* inotify descriptors */
	int inotify_fd;
	int inotify_wd_pts;

	/* global flags */
	int verbose;
	int client;
	int daemon;
	int trace_all;

	/* attached terminals */
	struct sf_terminal *terminal;
	int terminal_cnt;

	/* replay, values same as for terminal::display */
	int replay;
	char *replay_filename;
	int replay_fd;
	int replay_timeout;

	/* input */
/* user interface types */
#define INPUT_TYPE_CMD    1
#define INPUT_TYPE_MENU   2
	int input_type;
	int input_change_terminal;
	struct sf_terminal *terminal_current;
	int terminal_current_num;

#ifdef CONFIG_SNIFFY_X_RESIZE
	Display *display;
	Window term_window;
	int resize;
#endif
};


/* prototypes */
int sf_config_cmdline(struct sf_config *cfg, int argc, char **argv);
int sf_config_terminal_del(struct sf_config *cfg, struct sf_terminal *term);
struct sf_terminal* sf_config_terminal_add(struct sf_config *cfg, int idx);
struct sf_terminal* sf_config_terminal_find(struct sf_config *cfg, int idx);

int sf_fsm_terminal_process(struct sf_config *cfg, struct nl_msg *msg);

int sf_notify_init(struct sf_config *cfg);
int sf_notify_close(struct sf_config *cfg);
int sf_notify_process(struct sf_config *cfg);

int sf_log_print(struct sf_config *cfg, const char *str, ...);

int sf_trace_init(struct sf_config *cfg);
int sf_trace_close(struct sf_config *cfg);
int sf_trace_store(struct sf_config *cfg, struct sf_terminal *term, struct sniffy_msg_data *m);
int sf_trace_init_term(struct sf_config *cfg, struct sf_terminal *term);
int sf_trace_close_term(struct sf_config *cfg, struct sf_terminal *term);

int sf_input_init(struct sf_config *cfg, int attach_sf);
int sf_input_close(struct sf_config *cfg);
int sf_input_process(struct sf_config *cfg);
int sf_input_print(struct sf_config *cfg, char *str, ...);

int sf_input_cmd_process(struct sf_config *cfg, int c);
int sf_input_cmd_print(struct sf_config *cfg, char *str, va_list ap);

int sf_replay_idxinit(struct sf_config* cfg, struct sf_trace_header *header);
int sf_replay_process(struct sf_config *cfg);
struct sf_trace_record *sf_replay_next(struct sf_config *cfg);
struct sf_trace_record *sf_replay_prev(struct sf_config *cfg);

int sf_resize_init(struct sf_config *cfg);
int sf_resize_close(struct sf_config *cfg);
int sf_resize_process(struct sf_config *cfg, struct sniffy_msg_ws *ws);

int sf_daemon_init(struct sf_config *cfg);
int sf_daemon_process(struct sf_config *cfg);
int sf_daemon_client(struct sf_config *cfg);
int sf_daemon_close(struct sf_config *cfg);

int sf_chkadd_dir(struct sf_config *cfg, char *dir);
int sf_set_term(int set_init);

int sf_intf_init(struct sf_config *cfg);
int sf_intf_close(struct sf_config *cfg);
int sf_intf_fd(struct sf_config *cfg);
int sf_intf_attach(struct sf_config *cfg, struct sf_terminal *term, int detach);
int sf_intf_recv(struct sf_config *cfg);

int sf_term_add(struct sf_config *cfg, struct sf_terminal* term);
int sf_term_del(struct sf_config *cfg, struct sf_terminal* term, int detach);

int sniffy_process(struct sf_config *cfg, struct sf_terminal *term, struct sniffy_msg_data *msg);
void usage(void);

#define SF_LOG_PRINT1(cfg, str) \
do { \
	char buf[1024]; \
	sprintf(buf, "[%s:%-5d] %s", __FUNCTION__, __LINE__, str); \
	sf_log_print(cfg, buf); \
} while(0)

#define SF_LOG_PRINT2(cfg, str, ...) \
do { \
	char lpbuf[1024]; \
	sprintf(lpbuf, "[%s:%-5d] %s", __FUNCTION__, __LINE__, str); \
	sf_log_print(cfg, lpbuf, __VA_ARGS__); \
} while(0)


#endif // !SF_CONFIG_H
