
#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <stdlib.h>
#include <sys/select.h>


#include "config.h"

extern int exit_fl;


int sf_daemon_init(struct sf_config *cfg)
{
	char socket_name[1000];
	struct sockaddr_un addr;
	int fl;
	pid_t child;

	fprintf(cfg->tty_out, "sniffy v"SNIFFY_VER" daemonized\n");

	if (-1 == (child = fork())) {
		perror("fork failed");	
		return -1;
	}

	if (child) {
		SF_LOG_PRINT2(cfg, "spawned child pid = %d\n", child);
		_exit(0);
	}

	if (-1 == sf_chkadd_dir(cfg, cfg->sniffy_dir))
		return -1;

	sprintf(socket_name, "%s/din",  cfg->sniffy_dir);

	if (-1 == (cfg->input_fd = socket(PF_UNIX, SOCK_DGRAM, 0))) {
		perror("socket failed");
		return -1;
	}

        if (-1 == (fl = fcntl(cfg->input_fd, F_GETFL, 0)))
            return -1;

        if (-1 == fcntl(cfg->input_fd, F_SETFL, fl | O_NDELAY))
            return -1;

	addr.sun_family = AF_UNIX;
	strcpy(addr.sun_path, socket_name);

	if (-1 == bind(cfg->input_fd, (const struct sockaddr *) &addr, sizeof(addr))) {
		SF_LOG_PRINT2(cfg, "bind failed for %s\n", socket_name);
		perror("bind failed");
	        fprintf(cfg->tty_out, "another sniffy instance is probably running, or a file \"%s\" exists\n", 
			socket_name);
		return -1;
	}

	SF_LOG_PRINT2(cfg, "running as a daemon, with input socket %s\n", socket_name);

	if (NULL == (cfg->tty_out = fopen("/dev/null", "w"))) {
		perror("fopen failed");
		return 0;
	}

	setsid();
	return 0;
}

int sf_daemon_process(struct sf_config *cfg)
{
	struct sf_imsg msg;

	read(cfg->input_fd, &msg, sizeof(msg));

	SF_LOG_PRINT2(cfg, "got command %d\n", msg.cmd);

	if (SF_IMSG_ATTACH == msg.cmd) {
		char name[100];

		switch(msg.attach.type) {
		case SF_TYPE_TTY: 
			sprintf(name, "/dev/tty%d", msg.attach.idx);
			break;
		case SF_TYPE_TTYS: 
			sprintf(name, "/dev/ttyS%d", msg.attach.idx);
			break;
		case SF_TYPE_PTY: 
			sprintf(name, "/dev/pts/%d", msg.attach.idx);
			break;
		case SF_TYPE_CONSOLE: 
			sprintf(name, "/dev/console");
			break;
		}

		if (NULL == (cfg->tty_out = fopen(name, "w"))) {
			perror("fopen failed");
			return 0;
		}

		return 0;
	}

	if (SF_IMSG_DETACH == msg.cmd) {

		if (NULL == (cfg->tty_out = fopen("/dev/null", "w"))) {
			perror("fopen failed");
			return 0;
		}
		return 0;
	}

	if (SF_IMSG_DATA == msg.cmd)
		return msg.data;

	SF_LOG_PRINT2(cfg, "unknown command %d\n", msg.cmd);
	return 0;
}

int sf_daemon_close(struct sf_config *cfg)
{
	char socket_name[100];
	sprintf(socket_name, "%s/din",  cfg->sniffy_dir);

	close(cfg->input_fd);

	if (-1 == remove(socket_name)) {
		perror("remove failed");
		return -1;
	}

	SF_LOG_PRINT1(cfg, "socket removed\n");
	return 0;
}

int sf_daemon_client(struct sf_config *cfg)
{
	struct sockaddr_un addr;
	int fd;
	char *tty_name;
	int idx, type;
	struct sf_imsg msg;

	fprintf(cfg->tty_out, "sniffy v"SNIFFY_VER " client\n");

        if (-1 == (fd = socket(PF_UNIX, SOCK_DGRAM, 0))) {
		perror("socket failed");
		return -1;
	}

        addr.sun_family = AF_UNIX;
        sprintf(addr.sun_path, "%s/din", cfg->sniffy_dir);

	if (NULL == (tty_name = ttyname(0))) {
		perror("ttyname failed");
		return -1;
	}

	SF_LOG_PRINT2(cfg, "ttyname %s\n", tty_name);

	if (1 == sscanf(tty_name, "/dev/pts/%d", &idx))
		type = SF_TYPE_PTY;
	else if (1 == sscanf(tty_name, "/dev/tty%d", &idx))
		type = SF_TYPE_TTY;
	else if (1 == sscanf(tty_name, "/dev/ttyS%d", &idx))
		type = SF_TYPE_TTYS;
	else if (!strcmp(tty_name, "/dev/console"))
		type = SF_TYPE_CONSOLE;
	else {
		perror("sscanf failed");
		return -1;
	}

	msg.cmd = SF_IMSG_ATTACH;
	msg.attach.idx = idx;
	msg.attach.type = type;

	if (-1 == sendto(fd, &msg, sizeof(msg), 0, (const struct sockaddr *) &addr, sizeof(addr))) {
		perror("sendto failed");
		return -1;
	}

	SF_LOG_PRINT1(cfg, "atach sent\n");

	sf_set_term(0);

	while(!exit_fl) {
		fd_set rfds;
		struct timeval tv = { 1, 0 };

		FD_ZERO(&rfds);
		FD_SET(0, &rfds);

		if (-1 == select(1, &rfds, NULL, NULL, &tv))
			continue;

		if (!FD_ISSET(0, &rfds))
			continue;

		msg.cmd = SF_IMSG_DATA;
		msg.data = getchar();

		if (-1 == sendto(fd, &msg, sizeof(msg), 0, (const struct sockaddr *) &addr, sizeof(addr))) {
			perror("sendto failed");
			exit_fl = 1;
		}

		SF_LOG_PRINT2(cfg, "data sent %c\n", msg.data);

		if ('q' == msg.data)
			exit_fl = 1;
        }

	msg.cmd = SF_IMSG_DETACH;
	sendto(fd, &msg, sizeof(msg), 0, (const struct sockaddr *) &addr, sizeof(addr));

	close(fd);
	sf_set_term(1);

	fprintf(cfg->tty_out, "client detached\n");
	return 0;
}
