
#include <stdio.h>
#include <unistd.h>
#include <sys/select.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>

#include <netlink/netlink.h>
#include <netlink/types.h>
#include <netlink/utils.h>
#include <netlink/msg.h>

#include "isniffy.h"
#include "config.h"

static struct sf_config cfg;
int exit_fl = 0;

static void signal_handler(int s)
{
	SF_LOG_PRINT2(&cfg, "got signal: %d", s);
	exit_fl = 1;
}

static int attach_terminals(int detach)
{
	struct sf_terminal *term;

	for(term = cfg.terminal; term; term = term->next) {

		if ((!detach) && (-1 == sf_term_add(&cfg, term)))
			return -1;
		
		if ((detach) && (-1 == sf_term_del(&cfg, term, 1)))
			return -1;
	}

	return 0;
}

static int do_stats(struct sf_config *cfg, struct sf_terminal *term, int type, int len)
{
	term->rx_bytes += len;
	term->rx_count++;

	if (type == SNIFFY_MSG_DATA_SLAVE) {
		term->rx_bytes_pty_slave += len;
		term->rx_count_pty_slave++;
	}

	if (type == SNIFFY_MSG_DATA_MASTER) {
		term->rx_bytes_pty_master+= len;
		term->rx_count_pty_master++;
	}

	return 0;
}

int sniffy_process(struct sf_config *cfg, struct sf_terminal *term, struct sniffy_msg_data *msg)
{
	int   type = msg->type;
	char *data = msg->data;
	int   len  = msg->len;

	if (cfg->terminal_current == term) {
#ifdef CONFIG_SNIFFY_X_RESIZE
		struct sniffy_msg_ws *ws = &msg->ws;

		if (cfg->resize)
			sf_resize_process(cfg, ws);
#endif

		if ((type == SNIFFY_MSG_DATA_SLAVE) && (term->display_slave))
			fwrite(data, len, 1, cfg->tty_out);

		if ((type == SNIFFY_MSG_DATA_MASTER) && (term->display_master))
			fwrite(data, len, 1, cfg->tty_out);

		fflush(cfg->tty_out);
	}

	sf_trace_store(cfg, term, msg);
	do_stats(cfg, term, type, len);
	return 0;
}

static int dispatch()
{
	fd_set read_fds;
	int max_fd;

	if (sf_intf_init(&cfg))
		return -1;

	if (-1 == attach_terminals(0))
		return -1;

	/* setup permanent descriptors for select,
	   input descriptor might be changed */
	FD_ZERO(&read_fds);
	FD_SET(nl_socket_get_fd(cfg.nlh), &read_fds);
	FD_SET(cfg.inotify_fd, &read_fds);

#define sf_max(a, b) (a) < (b) ? (b) : (a)
	max_fd = sf_max(nl_socket_get_fd(cfg.nlh), cfg.inotify_fd);

	while(!exit_fl) {
		struct timeval tv = { 1, 0 };
		int ret;
		fd_set rfds = read_fds;

		/* ... input descriptor might be changed */
		if (-1 != cfg.input_fd)
			FD_SET(cfg.input_fd, &rfds);

		max_fd = sf_max(max_fd, cfg.input_fd);
		ret = select(max_fd + 1, &rfds, NULL, NULL, &tv);

		if (ret == -1) {
			SF_LOG_PRINT2(&cfg, "select failed: %s", strerror(errno));
			continue;
		} else if (!ret)
			continue;

		/* netlink socket */
		if (FD_ISSET(nl_socket_get_fd(cfg.nlh), &rfds)) {
			sf_intf_recv(&cfg);
			ret--;
		}

		/* inotify descriptor */
		if (ret && FD_ISSET(cfg.inotify_fd, &rfds)) {
			sf_notify_process(&cfg);
			ret--;
		}

		/* input descriptor */
		if (ret && FD_ISSET(cfg.input_fd, &rfds)) {
			sf_input_process(&cfg);
		}
	}

	attach_terminals(1);
	nl_handle_destroy(cfg.nlh);
	return 0;
}

int main(int argc, char **argv)
{
	jmp_buf env;
	int ret;

	if (-1 == sf_config_cmdline(&cfg, argc, argv))
		usage();

	signal(SIGTERM, signal_handler);
	signal(SIGINT, signal_handler);

#ifdef CONFIG_SNIFFY_X_RESIZE
	if (cfg.resize && (-1 == sf_resize_init(&cfg)))
		return -1;
#endif

	/* start our separate client applications, 
	   before the serious work */
	if (cfg.replay)
		return sf_replay_process(&cfg);

	if (cfg.client)
		return sf_daemon_client(&cfg);


	if (-1 == sf_notify_init(&cfg)) {
		perror("sf_notify_init failed");
		return -1;
	}
		
	if (cfg.daemon && (-1 == sf_daemon_init(&cfg)))
		return -1;

	/* longjmp(env, ret) - close program
	   ret = 1 -> sf_close
	   ret = 2 -> trace_close, sf_close
	   ret = 3 -> everything is closed */
	if ((ret = setjmp(env))) {
		if (ret > 2)
			sf_input_close(&cfg);

		if (ret > 1)
			sf_trace_close(&cfg);

		if (cfg.daemon)
			sf_daemon_close(&cfg);

		sf_notify_close(&cfg);
		return 0;
	}

	if (-1 == sf_trace_init(&cfg))
		longjmp(env, 1);

	if (-1 == (sf_input_init(&cfg, 1)))
		longjmp(env, 2);

	fprintf(cfg.tty_out, "sniffy v"SNIFFY_VER"\n");

	dispatch();

	fprintf(cfg.tty_out, "server finished\n");

	longjmp(env, 3);
}
