
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <setjmp.h>
#include <unistd.h>
#include <stdlib.h>
#include <mcheck.h>
#include <string.h>

#include "config.h"


#define REPLAY_INDEX_MAX  100
static struct sf_trace_record *replay_index[REPLAY_INDEX_MAX] __attribute__((section (".bss")));
static unsigned int replay_index_size = -1;
static unsigned int replay_index_pos;

#define OFFSET_STACK_SIZE 100
static unsigned int *offset_stack = NULL;
static long offset_stack_pos = 0;
static long offset_stack_size = 0;

static int offset_push(struct sf_config *cfg, unsigned int offset)
{
	if (0 == offset_stack_size) {
		if (NULL == (offset_stack = (unsigned int*) malloc(OFFSET_STACK_SIZE*sizeof(unsigned int)))) {
			perror("malloc failed1");
			return -1;
		}
		offset_stack_size = OFFSET_STACK_SIZE;
		SF_LOG_PRINT2(cfg, "initial offset stack size %u\n", offset_stack_size);
	}

	SF_LOG_PRINT2(cfg, "storing offset %u with offset_stack_pos %u\n", offset, offset_stack_pos);

	offset_stack[offset_stack_pos++] = offset;	

	if (offset_stack_pos == offset_stack_size) {
		if (NULL == (offset_stack = (unsigned int*) realloc(offset_stack, (offset_stack_size + OFFSET_STACK_SIZE)*sizeof(unsigned int)))) {
			perror("malloc failed2");
			return -1;
		}

		offset_stack_size += OFFSET_STACK_SIZE;
		SF_LOG_PRINT2(cfg, "new offset stack size %u\n", offset_stack_size, offset_stack);
	}

	SF_LOG_PRINT2(cfg, "EXIT offset_stack_pos = %u\n", offset_stack_pos);
	return 0;
}

static int offset_pop(struct sf_config *cfg, unsigned int *offset)
{
	SF_LOG_PRINT2(cfg, "ENTRY offset_stack_pos = %ld\n", offset_stack_pos);

	// first pop, then return the value 
	offset_stack_pos -= 2;

	SF_LOG_PRINT2(cfg, "offset_stack_pos = %ld\n", offset_stack_pos);

	if (offset_stack_pos < 0) {
		offset_stack_pos = 1;
		return -1;
	}

	*offset = offset_stack[offset_stack_pos];
	return 0;
}

static struct sf_trace_record* read_rec(struct sf_config *cfg)
{
	struct sf_trace_record rec_head;
	struct sf_trace_record *rec;
	size_t ret;
	jmp_buf env;

	setjmp(env);

	SF_LOG_PRINT1(cfg, "ENTRY\n");

	ret = read(cfg->replay_fd, &rec_head, sizeof(struct sf_trace_record));

	// EOF
	if (0 == ret)
		return NULL;
	// error
	if (-1 == ret) {
		perror("read failed");
		return NULL;
	}

	// got zero record, jump to page alligned offset for new data 
	if (!rec_head.used) {
		off_t cur;
		int page_size = sysconf(_SC_PAGESIZE);

		SF_LOG_PRINT1(cfg, "got ZERO timestamp\n");

		if (-1 == (cur = lseek(cfg->replay_fd, 0, SEEK_CUR))) {
			perror("lseek failed");
			return NULL;
		}

		cur -= sizeof(struct sf_trace_record);
		SF_LOG_PRINT2(cfg, "current offset: %u\n", cur);

		cur = (cur + page_size) & ~(page_size - 1);
		SF_LOG_PRINT2(cfg, "changing replay file offset to %lu\n", cur);

		if (-1 == (cur = lseek(cfg->replay_fd, cur, SEEK_SET))) {
			perror("lseek failed");
			return NULL;
		}

		longjmp(env, 0);
	}

	if (NULL == (rec = malloc(sizeof(struct sf_trace_record) + rec_head.len + 1))) {
		SF_LOG_PRINT2(cfg, "malloc failed: %u\n", sizeof(struct sf_trace_record) + rec_head.len + 1);
		perror("malloc failed3");
		return NULL;
	}

	*rec = rec_head;

	ret = read(cfg->replay_fd, rec->data, rec->len);
	if (0 == ret)
		return NULL;

	if (-1 == ret) {
		perror("read failed");
		return NULL;
	}

	SF_LOG_PRINT1(cfg, "EXIT\n");
	return rec;
}

static int index_cleanup(struct sf_config *cfg)
{
	int i;

	SF_LOG_PRINT1(cfg, "ENTRY\n");
	for(i = 0; i < REPLAY_INDEX_MAX; i++) {

		if (replay_index[i]) {
			free(replay_index[i]);
			SF_LOG_PRINT2(cfg, "clean %d %p\n", i, replay_index[i]);
			replay_index[i] = NULL;
		}
	}

	SF_LOG_PRINT1(cfg, "EXIT\n");

	return 0;
}

static int index_create(struct sf_config *cfg, unsigned int offset)
{
	int i;

	SF_LOG_PRINT2(cfg, "offset = %u\n", offset);

	if (-1 == index_cleanup(cfg))
		return -1;

	if (-1 == lseek(cfg->replay_fd, offset, SEEK_SET)) {
		perror("lseek failed");
		return -1;
	}

	offset_push(cfg, offset);

	for(i = 0; i < REPLAY_INDEX_MAX; i++) {
		struct sf_trace_record *rec;

		if (NULL == (rec = read_rec(cfg))) {
			i++;
			break;
		}

		replay_index[i] = rec;
	}

	replay_index_size = i;
	replay_index_pos = 0;
	return 0;
}

int sf_replay_idxinit(struct sf_config* cfg, struct sf_trace_header *header)
{
	printf("replay openning file %s\n", cfg->replay_filename);

	if (-1 == (cfg->replay_fd = open(cfg->replay_filename, O_RDONLY))) {
		perror("open failed");
		return -1;
	}

	if (-1 == read(cfg->replay_fd, header, sizeof(*header))) {
		perror("read failed");
		return -1;
	}

	if (-1 == index_create(cfg, sizeof(struct sf_trace_header)))
		return -1;

	return 0;
}

struct sf_trace_record *sf_replay_next(struct sf_config *cfg)
{
	struct sf_trace_record *rec;

	SF_LOG_PRINT2(cfg, "ENTRY replay_index_pos = %d, replay_index_size = %d\n", replay_index_pos, replay_index_size);

	if ((replay_index_pos + 1) == replay_index_size) {
		unsigned int offset;

		if (replay_index_size < REPLAY_INDEX_MAX)
			return NULL;

		if (-1 == (offset = lseek(cfg->replay_fd, 0, SEEK_CUR))) {
			perror("lseek failed");
			return NULL;
		}

		if (-1 == index_create(cfg, offset))
			return NULL;
	} else
		replay_index_pos++;

	rec = replay_index[replay_index_pos];

	SF_LOG_PRINT2(cfg, "EXIT replay_index_pos = %d\n", replay_index_pos);
	return rec;
}

struct sf_trace_record *sf_replay_prev(struct sf_config *cfg)
{
	unsigned int offset;
	jmp_buf env;
	int exit = 0;

	if (setjmp(env))
		exit = 1;

	SF_LOG_PRINT2(cfg, "ENTRY replay_index_pos = %d, replay_index_size = %d\n", replay_index_pos, replay_index_size);

	if (!replay_index_pos) {
		if (-1 == offset_pop(cfg, &offset)) {

			replay_index_pos = 1;

		} else {
			SF_LOG_PRINT2(cfg, "got offset %u\n", offset);

			if (-1 == index_create(cfg, offset))
				return NULL;

			replay_index_pos = replay_index_size;
		}
	}

	replay_index_pos--;

	if (!exit) {
		longjmp(env, 1);
	}

	SF_LOG_PRINT2(cfg, "EXIT replay_index_pos = %d\n", replay_index_pos);
	return replay_index[replay_index_pos];
}
