
#include <setjmp.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdio_ext.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <curses.h>
#include <term.h>
#include <time.h>

#define tparm3(a,b,c) tparm(a,b,c,0,0,0,0,0,0,0)

#ifdef CONFIG_SNIFFY_X_RESIZE
#include <X11/X.h>
#include <X11/Xlib.h>
#endif

#include "config.h"

static int display_master = 0;
static int display_slave  = 1;
static int display_status = 1;

static struct sf_trace_header header;
static char status_start[1000];

enum { 
	REPLAY_NONE = 0, 
	REPLAY_FIRST,
	REPLAY_PREV,
	REPLAY_NEXT,
	REPLAY_QUIT,
	REPLAY_TIMEOUT,
	REPLAY_SLAVE,
	REPLAY_MASTER,
	REPLAY_STATUS
};

int replay_input(struct sf_config *cfg, struct timeval *tv)
{
	int c, ret;
	jmp_buf env;
	fd_set rfds;

	setjmp(env);

	FD_ZERO(&rfds);
	FD_SET(0, &rfds);

	ret = select(1, &rfds, NULL, NULL, tv);

	/* select failed - finish replay */
	if (-1 == ret) {
		SF_LOG_PRINT1(cfg, "\nselect failed");
		return REPLAY_QUIT;
	} 

	/* timeout - continue with next record */
	if (!ret)
		return REPLAY_NEXT;

	/* this should not happen... anyway :) */
	if (!FD_ISSET(0, &rfds))
		return REPLAY_QUIT;

	/* we got user input */
	c = getchar();

	if (c == '\033') {
		__fpurge(stdin);
		longjmp(env, 1);
	}

	switch(c) {
	case 'p' : return REPLAY_PREV;
	case 'n' : return REPLAY_NEXT;
	case 'f' : return REPLAY_FIRST;
	case 'q' : return REPLAY_QUIT;
	case 't' : return REPLAY_TIMEOUT;
	case 's' : return REPLAY_SLAVE;
	case 'm' : return REPLAY_MASTER;
	case 'h' : return REPLAY_STATUS;
	}

	longjmp(env, 1);
	return 0;
}

static int replay_print(struct sf_config *cfg, struct sf_trace_record *rec)
{
#ifdef CONFIG_SNIFFY_X_RESIZE
	if (cfg->resize) {
		XResizeWindow(cfg->display, cfg->term_window, rec->ws.ws_xpixel, rec->ws.ws_ypixel);
		XSync(cfg->display, 0);
	}
#endif
	SF_LOG_PRINT2(cfg, "got len %d, type = %d\n", rec->len);

        if ((rec->type == SNIFFY_MSG_DATA_SLAVE) && (display_slave))
                fwrite(rec->data, rec->len, 1, stdout);

        if ((rec->type == SNIFFY_MSG_DATA_MASTER) && (display_master))
                fwrite(rec->data, rec->len, 1, stdout);

	fflush(NULL);

	return 0;
}

static struct timeval* get_timeout(struct sf_config *cfg, struct timeval *tv_new, struct timeval *tv_old)
{
	static struct timeval tv;

	SF_LOG_PRINT1(cfg, "entry\n");
	SF_LOG_PRINT2(cfg, "got old record time: %u %u\n", tv_old->tv_sec, tv_old->tv_usec);

	if (!tv_new)
		return NULL;

	SF_LOG_PRINT2(cfg, "got new record time: %u %u\n", tv_new->tv_sec, tv_new->tv_usec);

	tv.tv_sec = tv_new->tv_sec - tv_old->tv_sec;

	if (tv_new->tv_usec >= tv_old->tv_usec)
		tv.tv_usec = tv_new->tv_usec - tv_old->tv_usec;
	else {
		tv.tv_usec = tv_new->tv_usec + (1000000 - tv_old->tv_usec);
		if (tv.tv_sec) 
			tv.tv_sec--;
	}

	SF_LOG_PRINT2(cfg, "EXIT : %u %u\n\n", tv.tv_sec, tv.tv_usec);
	return &tv;
}

static int status_init(struct sf_config *cfg, struct sf_trace_header *header)
{
	struct tm *tm;

	setupterm((char *)0, 1, (int *)0);

	SF_LOG_PRINT2(cfg, "got init time: %u %u, maxidx %lu\n", header->time.tv_sec, header->time.tv_usec, header->maxidx);
	
	if (NULL == (tm = localtime(&header->time.tv_sec))) {
		perror("localtime failed");
		return -1;
	}

	sprintf(status_start, "%02d.%02d.%02d-%02d:%02d:%02d.%03lu",
			tm->tm_mday,
			tm->tm_mon,
			(tm->tm_year + 1900) % 100,
			tm->tm_hour,
			tm->tm_min,
			tm->tm_sec,
			header->time.tv_usec / 1000);

	return 0;
}

static int status_print(struct sf_config *cfg, struct sf_trace_record *rec)
{
	struct tm *tm;
	char status_rec[1000];
	float per;
	unsigned long maxidx = header.maxidx - 1;

	if (NULL == (tm = localtime(&rec->time.tv_sec))) {
		perror("localtime failed");
		return -1;
	}

        sprintf(status_rec, "%02d.%02d.%02d-%02d:%02d:%02d.%03lu",
			tm->tm_mday,
			tm->tm_mon,
			(tm->tm_year + 1900) % 100,
			tm->tm_hour,
			tm->tm_min,
			tm->tm_sec,
			rec->time.tv_usec / 1000);

	putp(save_cursor);
	putp(tparm3(cursor_address, 0, 0));
	putp(enter_reverse_mode);

	per = (float) 100 / maxidx * rec->idx;
	printf("%s - %s [%02d%% %lu/%lu]", status_start, status_rec, (int)per, rec->idx, maxidx);
	if (cfg->replay_timeout)
		printf(" T ");
	printf("       ");

	putp(enter_reverse_mode);
	putp(restore_cursor);
	fflush(NULL);
	return 0;
}

int sf_replay_process(struct sf_config *cfg)
{
	int ret;
	jmp_buf env;
	struct sf_trace_record *rec, *recn = NULL;
	int action = REPLAY_NEXT;

	if (-1 == (sf_input_init(cfg, 0)))
		return -1;

	if ((ret = setjmp(env))) {
		sf_input_close(cfg);
		return 0;
	}

	if (-1 == sf_replay_idxinit(cfg, &header))
		longjmp(env, 1);

	if (-1 == status_init(cfg, &header))
		longjmp(env, 1);

	rec = sf_replay_next(cfg);

	do {
		struct timeval *tv, tv_prev;

		replay_print(cfg, rec);

		if (display_status)
			status_print(cfg, rec);

		tv_prev = rec->time;
		recn = sf_replay_next(cfg);

		if (cfg->replay_timeout)
			tv = get_timeout(cfg, recn ? &recn->time : NULL, &tv_prev);
		else
			tv = NULL;

		action = replay_input(cfg, tv);

		switch(action) {
		case REPLAY_PREV: 
			recn = sf_replay_prev(cfg); 
			break;
		case REPLAY_TIMEOUT: 
			cfg->replay_timeout = !cfg->replay_timeout; 
			break;
		case REPLAY_SLAVE:  
			display_slave = !display_slave; 
			break;
		case REPLAY_MASTER: 
			display_master = !display_master;
			break;
		case REPLAY_STATUS: 
			display_status = !display_status;
			break;
		}

		if (recn)
			rec = recn;

	} while(REPLAY_QUIT != action);	

	close(cfg->replay_fd);

	printf("\nreplay finished\n");
	longjmp(env, 1);
}
