
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <time.h>

#include <autoconf.h>
#include "config.h"


#ifdef CONFIG_SNIFFY_TRACE_SIMPLE
static int trace_init_terminal_ll(struct sf_config *cfg, struct sf_terminal *term)
{
	return 0;
}

static int trace_close_terminal_ll(struct sf_config *cfg, struct sf_terminal *term)
{
	return 0;
}

static int trace_store_ll(struct sf_config *cfg, struct sf_terminal *term, struct sf_trace_record *rec, struct sniffy_msg_data *m)
{
	if (-1 == (write(term->trace_fd, rec, sizeof(*rec)))) {
		perror("write failed");
		return -1;
	}

	if (-1 == (write(term->trace_fd, m->data, rec->len))) {
		perror("write failed");
		return -1;
	}

	return 0;
}
#endif // CONFIG_SNIFFY_TRACE_SIMPLE


#ifdef CONFIG_SNIFFY_TRACE_MMAP

static int trace_mmap_open(struct sf_config *cfg, struct sf_terminal *term)
{
	int page_size = sysconf(_SC_PAGESIZE);
	int offset = term->trace_size & (~(page_size - 1));

	/* increase the file size and allign it to the page */
	term->trace_size += TRACE_MMAP_SIZE;
	term->trace_size &= ~(page_size - 1);

	SF_LOG_PRINT2(cfg, "open at offset %lu, mmap size %lu, total size %lu, fd %d\n", 
		offset, TRACE_MMAP_SIZE, term->trace_size, term->trace_fd);

	if (-1 == ftruncate(term->trace_fd, term->trace_size)) {
		perror("ftruncate failed");
		return -1;
	}

	if (MAP_FAILED == (term->trace_mmap_ptr = mmap(NULL, 
						TRACE_MMAP_SIZE, 
						PROT_WRITE | PROT_READ, 
						MAP_SHARED, 
						term->trace_fd, 
						offset))) {
		perror("mmap failed");
		return -1;
	}

	term->trace_mmap_cur = term->trace_mmap_ptr;

	/* skip the header on the first time */
	if (TRACE_MMAP_SIZE == term->trace_size)
		term->trace_mmap_cur += sizeof(struct sf_trace_header);
		

	SF_LOG_PRINT2(cfg, "ok, ptr %x\n", term->trace_mmap_cur);
	return 0;
}

static int trace_mmap_close(struct sf_config *cfg, struct sf_terminal *term)
{
	SF_LOG_PRINT2(cfg, "trace mmap - unmmap ptr %x\n", term->trace_mmap_ptr);

	if (-1 == munmap(term->trace_mmap_ptr, TRACE_MMAP_SIZE)) {
		perror("munmap failed");
		return -1;
	}

	return 0;
}

static int trace_init_terminal_ll(struct sf_config *cfg, struct sf_terminal *term)
{
	return trace_mmap_open(cfg, term);
}

static int trace_close_terminal_ll(struct sf_config *cfg, struct sf_terminal *term)
{
	int final_size;

	final_size = term->trace_size - TRACE_MMAP_SIZE + (term->trace_mmap_cur - term->trace_mmap_ptr);

	if (-1 == trace_mmap_close(cfg, term))
		return -1;

	SF_LOG_PRINT2(cfg, "trace mmap - truncating the file to size %lu\n", final_size);

	if (-1 == ftruncate(term->trace_fd, final_size)) {
		perror("ftruncate failed");
		return -1;
	}

	return 0;
}

static int trace_store_ll(struct sf_config *cfg, struct sf_terminal *term, struct sf_trace_record *rec, struct sniffy_msg_data *m)
{
	unsigned int free_space = (term->trace_mmap_ptr + TRACE_MMAP_SIZE) - term->trace_mmap_cur;
	unsigned int written = 0;

#define TRACE_MMAP_MINWRITE (sizeof(*rec) + 10)
#define TRACE_MMAP_2WRITE   (sizeof(*rec) + rec->len)

	/* check if we are going out of space */
	if (free_space < TRACE_MMAP_2WRITE) {

		/* write at least something */
		if (free_space > TRACE_MMAP_MINWRITE) {
			written = (free_space - sizeof(*rec));

			rec->len = written;
			memcpy(term->trace_mmap_cur, rec, sizeof(*rec));
			term->trace_mmap_cur += sizeof(*rec);
			memcpy(term->trace_mmap_cur, m->data, rec->len);
			term->trace_mmap_cur += rec->len;

			SF_LOG_PRINT2(cfg, "trace mmap - store last block data, size %lu, total size %lu\n", 
				TRACE_MMAP_2WRITE, term->trace_mmap_cur - term->trace_mmap_ptr);

			free_space = (term->trace_mmap_ptr + TRACE_MMAP_SIZE) - term->trace_mmap_cur;
		}

		if (free_space) {
			struct sf_trace_record *r = (struct sf_trace_record*) term->trace_mmap_cur;
			r->used = 0;
		}

		SF_LOG_PRINT2(cfg, "trace mmap - closing mmap with offset %x\n", 
					term->trace_size - TRACE_MMAP_SIZE + term->trace_mmap_cur - term->trace_mmap_ptr);

		if (trace_mmap_close(cfg, term))
			return -1;

		if (trace_mmap_open(cfg, term))
			return -1;
	}

	rec->len = m->len - written;
	memcpy(term->trace_mmap_cur, rec, sizeof(*rec));
	term->trace_mmap_cur += sizeof(*rec);
	memcpy(term->trace_mmap_cur, m->data + written, rec->len);
	term->trace_mmap_cur += rec->len;

	SF_LOG_PRINT2(cfg, "trace mmap - data stored, size %lu, total size %lu\n", 
		TRACE_MMAP_2WRITE, term->trace_mmap_cur - term->trace_mmap_ptr);

	return 0;
}
#endif // CONFIG_SNIFFY_TRACE_MMAP

int sf_chkadd_dir(struct sf_config *cfg, char *dir)
{
	struct stat st;

	if (-1 == stat(dir, &st)) {
		if (errno != ENOENT) {
			perror("stat failed");
			return -1;
		}

		if (-1 == mkdir(dir, S_IRUSR | S_IWUSR | S_IXUSR)) {
			perror("mkdir failed");
			return -1;
		}

		return 0;
	}

	if ((S_ISDIR(st.st_mode)) && 
	    (S_IRUSR & st.st_mode) &&
	    (S_IWUSR & st.st_mode))
		return 0;

	fprintf(cfg->tty_out, "check failed for directory %s\n", dir);
	return -1;
}

static int get_strdate(struct sf_config *cfg, char *buf)
{
	struct timeval tv;
	struct tm *tm;

	if (-1 == gettimeofday(&tv, NULL)) {
		perror("gettimeofday failed");
		return -1;
	}

	if (NULL == (tm = localtime(&tv.tv_sec))) {
		perror("localtime failed");
		return -1;
	}

	sprintf(buf, "%02d.%02d.%02d-%02d:%02d:%02d.%03lu",
			tm->tm_mday,
			tm->tm_mon,
			(tm->tm_year + 1900) % 100,
			tm->tm_hour,
			tm->tm_min,
			tm->tm_sec,
			tv.tv_usec / 1000);

	return 0;
}

static int set_filename(struct sf_config *cfg, struct sf_terminal *term)
{
#define MAXLEN 255
	char buff[MAXLEN];
	char buf[MAXLEN];

	if (-1 == get_strdate(cfg, buff))
		return -1;

	sprintf(buf, "%s --- %s", term->trace_filename, buff);

	if (-1 == rename(term->trace_filename, buf)) {
		perror("rename failed");
		return -1;
	}

	SF_LOG_PRINT2(cfg, "set filename %s\n", buf);
	return 0;
}

static char *get_filename(struct sf_config *cfg, struct sf_terminal *term)
{
	char bufd[MAXLEN];
	char buff[MAXLEN];
	char buf[MAXLEN];
	char *ret;
	int len;

	sprintf(bufd, "%s/%03d", cfg->sniffy_dir, term->idx);

	if (-1 == sf_chkadd_dir(cfg, bufd))
		return NULL;

	if (-1 == get_strdate(cfg, buff))
		return NULL;

	sprintf(buf, "%s/%s", bufd, buff);

	SF_LOG_PRINT2(cfg, "get filename [%s]\n", buf);

	len = strlen(buf);

	if (NULL == (ret = malloc(len + 1))) {
		perror("malloc failed");
		return NULL;
	}

	strncpy(ret, buf, len);
	ret[len] = 0x0;

	return ret;
}

static int trace_write_header(struct sf_config *cfg, struct sf_terminal *term)
{
	struct sf_trace_header header;
	struct timeval tv;
	ssize_t size;

	if (-1 == gettimeofday(&tv, NULL)) {
		perror("gettimeofday failed");
		return -1;
	}

	header.time = tv;

	if (-1 == (size = write(term->trace_fd, &header, sizeof(header)))) {
		perror("write failed");
		return -1;
	}

	term->trace_size = size;
	return 0;
}

int sf_trace_init_term(struct sf_config *cfg, struct sf_terminal *term)
{
	if (!term->trace)
		return 0;

	if (NULL == (term->trace_filename = get_filename(cfg, term)))
		return -1;

	if (-1 == (term->trace_fd = open(term->trace_filename, 
					O_CREAT | O_TRUNC | O_RDWR, 
					S_IRUSR | S_IWUSR))) {
		perror("open failed");
		return -1;
	}

	if (-1 == trace_write_header(cfg, term))
		return -1;

	if (-1 == trace_init_terminal_ll(cfg, term))
		return -1;

	return 0;
}

static int term_write_idx(struct sf_config *cfg, struct sf_terminal *term)
{
	struct sf_trace_header header;

	if (0 != lseek(term->trace_fd, 0, SEEK_SET)) {
		perror("lseek failed");
		return -1;
	}

	if (sizeof(header) != read(term->trace_fd, &header, sizeof(header))) {
		perror("read failed");
		return -1;
	}

	header.maxidx = term->trace_idx;

	if (0 != lseek(term->trace_fd, 0, SEEK_SET)) {
		perror("lseek failed");
		return -1;
	}

	if (sizeof(header) != write(term->trace_fd, &header, sizeof(header))) {
		perror("read failed");
		return -1;
	}

	return 0;
}

int sf_trace_close_term(struct sf_config *cfg, struct sf_terminal *term)
{
	if (!term->trace)
		return 0;

	trace_close_terminal_ll(cfg, term);
	term_write_idx(cfg, term);

	close(term->trace_fd);

	if (-1 == set_filename(cfg, term))
		SF_LOG_PRINT2(cfg, "failed to set final filename for sf_terminal %d\n", term->idx);

	return 0;
}

int sf_trace_init(struct sf_config *cfg)
{
	return sf_chkadd_dir(cfg, cfg->sniffy_dir);
}

int sf_trace_close(struct sf_config *cfg)
{
	return 0;
}

int sf_trace_store(struct sf_config *cfg, struct sf_terminal *term, struct sniffy_msg_data *m)
{
	struct sf_trace_record rec;
	struct timeval tv;

	if (!term->trace)
		return 0;

	if (-1 == gettimeofday(&tv, NULL)) {
		perror("gettimeofday failed");
		return -1;
	}

	rec.used = 1;
	rec.time = tv;
	rec.len  = m->len;
	rec.type = m->type;
	rec.ws   = m->ws;
	rec.idx  = term->trace_idx++;

	if (-1 == trace_store_ll(cfg, term, &rec, m))
		return -1;

	return 0;
}
